/*
   NeoPixel Cube Program

   This program does the following:
     Displays the string "Welcome to Craig & Heather's WoodHaven"
     Displays the time and date based on NTP and Timezone library
     Displays a current weather summary data provided by OpenWeatherMap.com
     and numerous interspersed colorful patterns

   NOTE: there are real issues with colors between text display on the matrix
         and colors for the NeoPixels. The matrix required 16 bit colors
         for the text whereas the NeoPixels require 32 bit colors. See:
         https://learn.adafruit.com/adafruit-neopixel-uberguide/neomatrix-library
         at the end of the page for the details. If the color types are confused
         strange things will happen.

   Version 1.1 - Reduced use of String types because I think they weren't being cleaned
   up correctly resulting in low memory conditions. More still may need to be done
   if problems persist.

   Version 1.2 - Added feels like temp and tightened error handling. Also changed
   weather data fetch time to 10 minutes.

   Version 1.3 - Added four new patterns

   Version 1.4 - Added more new patterns and reorganized the code
   to make it more managable. Also changed weather data fetch time to 30 minutes.

   Version 1.5 - Added lava pattern along with misc tweeks for stability but
   still kind of unstable. May have bad hardware. Also changed display delay
   from 5 to 10 minutes.

   Concept, Design and Implementation by: Craig A. Lindley
   Version: 1.5
   Last Update: 09/10/2020
*/

#include <Adafruit_GFX.h>
#include <Adafruit_NeoMatrix.h>
#include <Adafruit_NeoPixel.h>
#include <Time.h>
#include <Timezone.h>
#include <WiFiManager.h>
#include <FastLED.h>

#include "Colors.h"
#include "Patterns.h"
#include "NewPatterns1.h"
#include "NewPatterns2.h"
#include "NTP.h"
#include "OpenWeatherMap.h"

// Total number of patterns
#define NUMBER_PATTERNS 13

// Text string to display
#define TEXT_STRING "Welcome to Craig & Heather's Wood Haven"

// Colors for textual display
// Currently all the same
#define TEXT_COLOR    matrix.Color(128, 128, 128)
#define TIME_COLOR    matrix.Color(128, 128, 128)
#define WEATHER_COLOR matrix.Color(128, 128, 128)

// My OpenWeatherMap key
#define OWM_KEY "512f63fc90533134388edda77e7beaf5"

// Zip code for our house in Colorado Springs
#define OWM_ZIP "80919,us"

// Time delay between fetching weather conditions
#define DELAY_MINUTES 30
#define DELAY_SECONDS (DELAY_MINUTES * 60)
#define DELAY_MILLISECONDS (DELAY_SECONDS * 1000)

// Time to wait after weather before starting display sequence over
#define WAIT_MINUTES 10
#define WAIT_SECONDS (WAIT_MINUTES * 60)
#define WAIT_MILLISECONDS (WAIT_SECONDS * 1000)

// Pin on ESP8266 that drives the NeoPixels
#define PIN D1

// Name of AP when configuring WiFi credentials
#define AP_NAME "CubeAP"

// Checks WiFi every minute. Reboot if connection down
#define WIFI_TIMEOUT 60000

// This clock runs in the Mountain Time Zone
// Change this for your timezone
#define DST_TIMEZONE_OFFSET -6    // Day Light Saving Time offset (-6 is mountain time)
#define  ST_TIMEZONE_OFFSET -7    // Standard Time offset (-7 is mountain time)

// TimeZone and Daylight Savings Time Rules
// Define daylight savings time rules for the Mountain Time Zone
TimeChangeRule myDST = {"MDT", Second, Sun, Mar, 2, DST_TIMEZONE_OFFSET * 60};
TimeChangeRule mySTD = {"MST", First,  Sun, Nov, 2,  ST_TIMEZONE_OFFSET * 60};
Timezone myTZ(myDST, mySTD);

/***********************************************************************/
/*                          Driver Instantiations                      */
/***********************************************************************/

// Instantiate the NeoPixel matrix driver
// Cube is made of six 8x8 NeoPixel panels arranged as one row.
// Constructor below conveys this organization to the matrix library.
// See the NeoMatrix library for the details.
Adafruit_NeoMatrix matrix = Adafruit_NeoMatrix(
                              8, 8, 6, 1, PIN,
                              NEO_MATRIX_TOP  + NEO_MATRIX_LEFT +
                              NEO_MATRIX_ROWS + NEO_MATRIX_PROGRESSIVE +

                              NEO_TILE_TOP    + NEO_TILE_LEFT +
                              NEO_TILE_ROWS   + NEO_TILE_PROGRESSIVE,
                              NEO_GRB         + NEO_KHZ800);

// Instantiate the object that will retrieve the weather data
// Passing the API key
OWMCurrentWeather client(OWM_KEY, OWM_ZIP);

/***********************************************************************/
/*                               Variables                             */
/***********************************************************************/

unsigned long last_wifi_check_time;
TimeChangeRule *tcr;

// Data structure for holding Open Weather Map weather data
OWMCurrentData data;

// FSM states and state variable
enum STATES {WAKE, TEXT, TIMEDATE, WEATHER, SLEEP};
enum STATES state;

/***********************************************************************/
/*                            Misc Functions                           */
/***********************************************************************/

// Check to make sure WiFi is still connected
void wifiCheck(void) {

  uint32_t warning = Adafruit_NeoPixel::Color(255, 0, 0);

  unsigned long now = millis();

  if (now - last_wifi_check_time > WIFI_TIMEOUT) {
    Serial.print("Checking WiFi... ");
    if (WiFi.status() != WL_CONNECTED) {
      //      Serial.println("WiFi connection lost. Resetting...");
      matrix.clear();

      int loops = 20;

      // We have lost connection so indicate it
      while (loops-- > 0) {
        matrix.setPixelColor(FACE_LEFT_OFFSET  + 26, warning);
        matrix.setPixelColor(FACE_FRONT_OFFSET + 26, warning);
        matrix.setPixelColor(FACE_RIGHT_OFFSET + 26, warning);
        matrix.setPixelColor(FACE_BACK_OFFSET  + 26, warning);
        matrix.show();
        delay(1000);
        matrix.clear();
        matrix.show();
        delay(1000);
      }
      // Reset the esp8266
      ESP.reset();
    } else {
      Serial.println("WiFi OK");
    }
    last_wifi_check_time = now;
  } else {
    Serial.println("No WiFi check");
  }
}

// Display a text string on the cube. Text scrolls over all faces
// of the cube one time.
void displayText(const char *text, uint32_t color) {

  int16_t cursor, x1, y1;
  uint16_t h, textStringWidth;

  // Get the width in pixels of the text string to be displayed
  matrix.getTextBounds(text, 0, 0, &x1, &y1, &textStringWidth, &h);

  // Initialize variables
  cursor = matrix.width();

  // Set the color the text should be displayed with
  matrix.setTextColor(color);

  while (true) {
    yield();
    matrix.fillScreen(0);
    matrix.setCursor(cursor, 0);
    matrix.print(text);

    // The conditional below allows the text to be scrolled completely
    // off the matrix before returning.

    if (--cursor < -textStringWidth) {
      return;
    }
    matrix.show();
    delay(100);
  }
}

// Function to randomly pick and run a pattern
void pickPattern(void) {

  _pickPattern(random8(NUMBER_PATTERNS));
}

// Function to select and run a pattern
void _pickPattern(int selector) {

  Serial.print("Pattern: ");
  Serial.println(selector);

  switch (selector) {
    case 0:
      rainbow(1, FAST);
      delay(1000);
      colorWipe(0, FORWARD, MEDIUM);
      break;

    case 1:
      allRandom();
      delay(1000);
      dissolve(15, 100, MEDIUM);
      break;

    case 2:
      lightning(0, 15, 50, MEDIUM);
      break;

    case 3:
      lightning(Adafruit_NeoPixel::Color(255, 255, 255), 20, 50, MEDIUM);
      break;

    case 4:
      theaterChase(randomColor(), 12, SLOW);
      break;

    case 5:
      theaterChaseRainbow(1, MEDIUM);
      break;

    case 6:
      rainbow(1, FAST);
      delay(1000);
      colorWipe(0, BACKWARD, FAST);
      break;

    case 7:
      doPlasma(300);
      break;

    case 8:
      spin(false);
      break;

    case 9:
      spin(true);
      break;

    case 10:
      spinners(false);
      break;

    case 11:
      spinners(true);
      break;

    default:
    case 12:
      lava(5);
      break;
  }
  yield();
  // Turn off all NeoPixels after pattern has run
  matrix.clear();
  matrix.show();
}

/***********************************************************************/
/*                           Program Setup                             */
/***********************************************************************/

void setup() {

  Serial.begin(115200);
  delay(1000);

  // Seed the random number generators
  randomSeed(analogRead(A0));
  random16_add_entropy(time(0));

  // Initialize the cube for text/graphics display
  matrix.begin();
  matrix.clear();
  matrix.show();
  matrix.setTextWrap(false);

  // See notes in header about color differences
  matrix.setBrightness(30);
  matrix.Adafruit_NeoPixel::setBrightness(30);

  /*
    // Run all patterns on startup
    for (int i = 0; i < NUMBER_PATTERNS; i++) {
      _pickPattern(i);
      // Allow power supply to recharge
      delay(3000);
    }
  */

  // Local intialization. Once the WiFiManager's business is done,
  // there is no need to keep it around
  WiFiManager wifiManager;

  // Set configuration timeout to 5 minutes
  // If timeout occurs autoConnect will return but
  // without WiFi being connected.

  wifiManager.setConfigPortalTimeout(5 * 60);

  // Reset saved settings for testing purposes
  // Should be commented out for normal operation
  // wifiManager.resetSettings();

  // Fetches ssid and pass from eeprom and tries to connect
  // If it does not connect it starts an access point with the specified name
  // and goes into a blocking loop awaiting configuration
  wifiManager.autoConnect(AP_NAME);

  // Make sure WiFi connected
  if (WiFi.status() != WL_CONNECTED) {
    // Wait 10 minutes and restart
    delay(1000 * 60 * 10);
    ESP.reset();
  }

  // Initialize NTP code
  initNTP();

  // Initialize FSM state
  state = WAKE;
}

/***********************************************************************/
/*                           Program Loop                              */
/***********************************************************************/
// Buffer for formatting time, date and weather conditions
char buffer[100];

// Temporary storage for month string
char monthBuffer[5];

unsigned long nextUpdateTime, delayTime;

void loop() {

  yield();

  //  Serial.println(ESP.getFreeHeap(), DEC);

  // Do WiFi check
  wifiCheck();

  // Read the current UTC time from the NTP provider
  time_t utc = now();

  // Convert to local time taking DST into consideration
  time_t localTime = myTZ.toLocal(utc, &tcr);

  int hours = hour(localTime);

  // Is it after normal display hours ?
  // Clock should appear off between midnight and 6 AM
  if ((hours >= 0) && (hours < 6)) {
    matrix.clear();
    matrix.show();
    return;
  }

  // Run FSM
  switch (state) {

    case WAKE:

      Serial.println("\n\nState: WAKE\n");

      // Pick and run a random pattern
      pickPattern();

      // Clear shared buffers. Probably not needed.
      memset(buffer, 0, sizeof(buffer));
      memset(monthBuffer, 0, sizeof(monthBuffer));

      // Advance to next state
      state = TEXT;
      break;

    case TEXT:

      Serial.println("State: TEXT\n");

      // Display the text string
      displayText(TEXT_STRING, TEXT_COLOR);

      // Pick and run a random pattern
      pickPattern();

      // Advance to next state
      state = TIMEDATE;
      break;

    case TIMEDATE:

      Serial.println("State: TIMEDATE\n");

      strcpy(monthBuffer, monthShortStr(month(localTime)));

      sprintf(buffer, "%d:%02d%s %s %s %d %d",
              hourFormat12(localTime),
              minute(localTime),
              isPM(localTime) ? "pm" : "am",

              dayShortStr(weekday(localTime)),
              monthBuffer,
              day(localTime),
              year(localTime));

      // Display the text string
      displayText(buffer, TIME_COLOR);

      // Pick and run a random pattern
      pickPattern();

      // Advance to next state
      state = WEATHER;
      break;

    case WEATHER:

      Serial.println("State: WEATHER\n");

      // Is it time to fetch the current weather conditions
      if (millis() > nextUpdateTime) {
        nextUpdateTime = millis() + DELAY_MILLISECONDS;

        // Clear all previous weather data
        memset((void *) &data, 0, sizeof(OWMCurrentData));

        // Attempt to retrieve the current weather conditions
        // using zip code
        client.getCurrentWeatherData(&data);
      }

      // Build a string for display of weather data
      sprintf(buffer, "%s %.1ff Wind: %.1fmph Hum: %d%%",
              data.main, data.temp, data.windSpeed, data.humidity);

      // Display the text string
      displayText(buffer, WEATHER_COLOR);

      // Pick and run a random pattern
      pickPattern();

      // Advance to next state
      state = SLEEP;

      // Setup for display sequence delay
      delayTime = WAIT_MILLISECONDS + millis();
      break;

    case SLEEP:

      Serial.println("State: SLEEP\n");
      if (millis() > delayTime) {
        state = WAKE;
      } else {
        delay(10000);
      }
      break;
  }
}
