/*
   New Pattern Functions

   Concept, Design and Implementation: Craig A. Lindley
   Last Update: 08/26/2020
*/

#ifndef NEW_PATTERNS_1_H
#define NEW_PATTERNS_1_H

extern Adafruit_NeoMatrix matrix;

byte PERIMETER_LEDs [] {
  0,  1,  2,  3,  4,  5,  6, 7,
  15, 23, 31, 39, 47, 55, 63,
  62, 61, 60, 59, 58, 57, 56,
  48, 40, 32, 24, 16,  8
};
#define NUM_OF_PERIMETER_LEDS 28

int FACE_INDEX [] {
  FACE_LEFT_OFFSET,
  FACE_FRONT_OFFSET,
  FACE_RIGHT_OFFSET,
  FACE_BACK_OFFSET
};
#define NUM_OF_FACES 4

void _spin(int delayMS, boolean wipe) {

  const int faceColorIncrement = 9;
  const int tbColorIncrement   = 9;

  int faceColorIndex = random(256);
  int tbColorIndex = random(256);

  int faceIndex  = 0;
  int topIndexA = 0;
  int topIndexB = 14;
  int bottomIndexA = 14;
  int bottomIndexB = 0;
  int rowIndex   = 7;
  int colIndex   = 0;

  boolean done = false;
  while (! done) {
    yield();

    // Conditionally clear previous LEDs
    if (wipe) {
      matrix.clear();
    }

    // Next set the LED on the face
    int ledIndex = FACE_INDEX[faceIndex] + (8 * rowIndex) + colIndex;
    matrix.setPixelColor(ledIndex, Wheel(faceColorIndex));

    // Advance color selection
    faceColorIndex += faceColorIncrement;
    faceColorIndex %= 256;

    colIndex++;
    if (colIndex >= 8) {
      colIndex = 0;

      faceIndex++;
      if (faceIndex >= NUM_OF_FACES) {
        faceIndex = 0;

        rowIndex--;
        if (rowIndex == -1) {
          matrix.clear();

          done = true;
        }
      }
    }
    yield();

    // Pick colors for the opposing LEDs on top and bottom
    uint32 colorA = Wheel(tbColorIndex);
    uint32 colorB = Wheel(255 - tbColorIndex);

    // Next set the parameter LEDs on the top
    matrix.setPixelColor(PERIMETER_LEDs[topIndexA], colorA);
    matrix.setPixelColor(PERIMETER_LEDs[topIndexB], colorB);

    topIndexA += 1;
    if (topIndexA >= NUM_OF_PERIMETER_LEDS) {
      topIndexA = 0;
    }
    topIndexB += 1;
    if (topIndexB >= NUM_OF_PERIMETER_LEDS) {
      topIndexB = 0;
    }

    // Finally set the parameter LEDs on the bottom
    matrix.setPixelColor(FACE_BOTTOM_OFFSET + PERIMETER_LEDs[bottomIndexA], colorB);
    matrix.setPixelColor(FACE_BOTTOM_OFFSET + PERIMETER_LEDs[bottomIndexB], colorA);

    bottomIndexA -= 1;
    if (bottomIndexA < 0) {
      bottomIndexA = NUM_OF_PERIMETER_LEDS - 1;
    }
    bottomIndexB -= 1;
    if (bottomIndexB < 0) {
      bottomIndexB = NUM_OF_PERIMETER_LEDS - 1;
    }

    // Advance color selection
    tbColorIndex += tbColorIncrement;
    if (tbColorIndex > 255) {
      tbColorIndex = 0;
    }
    // Make changes visible
    matrix.show();
    delay(delayMS);
  }
}

// Spin pattern
void spin(boolean wipe) {

  matrix.clear();

  int delayMS = 50;
  for (int i = 0; i < 3; i++) {
    _spin(delayMS, wipe);
    delayMS >>= 1;
  }
}

#define topLoIndex       0
#define topHiIndex      40
#define frontLoIndex    41
#define frontHiIndex    80
#define rightLoIndex    81
#define rightHiIndex   120
#define backLoIndex    121
#define backHiIndex    160
#define leftLoIndex    161
#define leftHiIndex    200
#define bottomLoIndex  201
#define bottomHiIndex  240

typedef struct {
  int lo;
  int hi;
} RANGE;

RANGE RANGES [] {
  topLoIndex,    topHiIndex,
  frontLoIndex,  frontHiIndex,
  rightLoIndex,  rightHiIndex,
  backLoIndex,   backHiIndex,
  leftLoIndex,   leftHiIndex,
  bottomLoIndex, bottomHiIndex
};
#define NUM_OF_RANGES 6

int prevLedIndex = 0;

void doLED(int faceOffset, int ledIndex, int rangeSelectIndex, int *colorIndexPtr) {

  yield();

  int colorLoIndex = RANGES[rangeSelectIndex].lo;
  int colorHiIndex = RANGES[rangeSelectIndex].hi;

  if (*colorIndexPtr == -1) {
    *colorIndexPtr = colorLoIndex;
  }

  uint32_t color = Wheel(*colorIndexPtr);

  if (ledIndex % 2 == 0) {
    *colorIndexPtr += 2;
  } else  {
    *colorIndexPtr += 1;
  }

  matrix.setPixelColor(faceOffset + PERIMETER_LEDs[ledIndex], color);
  if (*colorIndexPtr > colorHiIndex) {
    *colorIndexPtr = colorLoIndex;
  }
}

void _spinners(int delayMS, boolean wipe) {

  boolean done = false;
  int topIndex    = -1;
  int frontIndex  = -1;
  int rightIndex  = -1;
  int backIndex   = -1;
  int leftIndex   = -1;
  int bottomIndex = -1;

  int rangeSelectorIndex;
  int shift = 0;

  while (! done) {

    for (int i = 0; i < NUM_OF_PERIMETER_LEDS; i++) {
      yield();

      // Top panel
      switch (shift) {
        case 0: rangeSelectorIndex = 0; break;
        case 1: rangeSelectorIndex = 1; break;
        case 2: rangeSelectorIndex = 2; break;
        case 3: rangeSelectorIndex = 3; break;
        case 4: rangeSelectorIndex = 4; break;
        case 5: rangeSelectorIndex = 5; break;
      }
      doLED(FACE_TOP_OFFSET, i, rangeSelectorIndex, &topIndex);

      // Front panel
      switch (shift) {
        case 0: rangeSelectorIndex = 5; break;
        case 1: rangeSelectorIndex = 0; break;
        case 2: rangeSelectorIndex = 1; break;
        case 3: rangeSelectorIndex = 2; break;
        case 4: rangeSelectorIndex = 3; break;
        case 5: rangeSelectorIndex = 4; break;
      }
      doLED(FACE_FRONT_OFFSET, i, rangeSelectorIndex, &frontIndex);

      // Right panel
      switch (shift) {
        case 0: rangeSelectorIndex = 4; break;
        case 1: rangeSelectorIndex = 5; break;
        case 2: rangeSelectorIndex = 0; break;
        case 3: rangeSelectorIndex = 1; break;
        case 4: rangeSelectorIndex = 2; break;
        case 5: rangeSelectorIndex = 3; break;
      }
      doLED(FACE_RIGHT_OFFSET, i, rangeSelectorIndex, &rightIndex);

      // Back panel
      switch (shift) {
        case 0: rangeSelectorIndex = 3; break;
        case 1: rangeSelectorIndex = 4; break;
        case 2: rangeSelectorIndex = 5; break;
        case 3: rangeSelectorIndex = 0; break;
        case 4: rangeSelectorIndex = 1; break;
        case 5: rangeSelectorIndex = 2; break;
      }
      doLED(FACE_BACK_OFFSET, i, rangeSelectorIndex, &backIndex);

      // Left panel
      switch (shift) {
        case 0: rangeSelectorIndex = 2; break;
        case 1: rangeSelectorIndex = 3; break;
        case 2: rangeSelectorIndex = 4; break;
        case 3: rangeSelectorIndex = 5; break;
        case 4: rangeSelectorIndex = 0; break;
        case 5: rangeSelectorIndex = 1; break;
      }
      doLED(FACE_LEFT_OFFSET, i, rangeSelectorIndex, &leftIndex);

      // Bottom panel
      switch (shift) {
        case 0: rangeSelectorIndex = 1; break;
        case 1: rangeSelectorIndex = 2; break;
        case 2: rangeSelectorIndex = 3; break;
        case 3: rangeSelectorIndex = 4; break;
        case 4: rangeSelectorIndex = 5; break;
        case 5: rangeSelectorIndex = 0; break;
      }
      doLED(FACE_BOTTOM_OFFSET, i, rangeSelectorIndex, &bottomIndex);

      matrix.show();
      delay(delayMS);

      if (wipe) {
        matrix.clear();
      }
    }
    topIndex    = -1;
    frontIndex  = -1;
    rightIndex  = -1;
    backIndex   = -1;
    leftIndex   = -1;
    bottomIndex = -1;

    shift++;
    if (shift == 6) {
      done = true;
    }
    matrix.clear();
  }
}

// Spinner pattern
void spinners(boolean wipe) {

  matrix.clear();

  int delayMS = 50;
  for (int i = 0; i < 3; i++) {
    _spinners(delayMS, wipe);
    delayMS >>= 1;
  }
}

#endif
