/*
   Misc Pattern Functions

   Concept, Design and Implementation: Craig A. Lindley
   Last Update: 08/29/2020
*/

#ifndef PATTERNS_H
#define PATTERNS_H

extern Adafruit_NeoMatrix matrix;

// Effect control parameters
#define FORWARD    0
#define BACKWARD   1
#define SLOW     250
#define MEDIUM    50
#define FAST       5

// LED offsets for each cube face
#define FACE_TOP_OFFSET      0
#define FACE_FRONT_OFFSET   64
#define FACE_RIGHT_OFFSET  128
#define FACE_BACK_OFFSET   192
#define FACE_LEFT_OFFSET   256
#define FACE_BOTTOM_OFFSET 320

// The total number of NeoPixel LEDs that make up the cube
#define NUM_LEDS (6 * 64)

CRGBPalette16 currentPalette;

// Set all NeoPixels to specified color
void allColor(uint32_t color) {
  for (int i = 0; i < NUM_LEDS; i++) {
    matrix.setPixelColor(i, color);
  }
  yield();
  matrix.show();
}

// Set all NeoPixels to a random color
void allRandom() {
  for (int i = 0; i < NUM_LEDS; i++) {
    matrix.setPixelColor(i, randomColor());
  }
  yield();
  matrix.show();
}

// Random dissolve colors
void dissolve(int simultaneous, int cycles, int speed) {
  for (int i = 0; i < cycles; i++) {
    yield();
    for (int j = 0; j < simultaneous; j++) {
      int idx = random(NUM_LEDS);
      matrix.setPixelColor(idx, 0);
    }
    matrix.show();
    delay(speed);
  }
  allColor(0);
}

// Wipes color from end to end
void colorWipe(uint32_t color, int speed, int direction) {
  for (int i = 0; i < NUM_LEDS; i++) {
    if (direction == FORWARD) {
      matrix.setPixelColor(i, color);
    }
    else {
      matrix.setPixelColor(NUM_LEDS - 1 - i, color);
    }
    matrix.show();
    delay(speed);
  }
}

// Rainbow colors that slowly cycle across LEDs
void rainbow(int cycles, int speed) {
  for (int j = 0; j < 256 * cycles; j++) {
    yield();
    for (int i = 0; i < NUM_LEDS; i++) {
      matrix.setPixelColor(i, Wheel(((i * 256 / NUM_LEDS) + j) & 255));
    }
    matrix.show();
    delay(speed);
  }
}

// Theater-style crawling lights
void theaterChase(uint32_t color, int cycles, int speed) {

  for (int j = 0; j < cycles; j++) {
    for (int q = 0; q < 3; q++) {
      for (int i = 0; i < NUM_LEDS; i = i + 3) {
        int pos = i + q;
        yield();
        matrix.setPixelColor(pos, color);
      }
      matrix.show();

      delay(speed);

      for (int i = 0; i < NUM_LEDS; i = i + 3) {
        matrix.setPixelColor(i + q, 0);
      }
    }
  }
}

// Theater-style crawling lights with rainbow effect
void theaterChaseRainbow(int cycles, int speed) { // TODO direction, duration
  for (int j = 0; j < 128 * cycles; j++) {   // cycle all 256 colors in the wheel
    for (int q = 0; q < 3; q++) {
      for (int i = 0; i < NUM_LEDS; i = i + 3) {
        int pos = i + q;
        yield();
        matrix.setPixelColor(pos, Wheel((i + j) % 255));
      }
      matrix.show();

      delay(speed);

      for (int i = 0; i < NUM_LEDS; i = i + 3) {
        yield();
        matrix.setPixelColor(i + q, 0);
      }
    }
  }
}

// Random flashes of lightning
void lightning(uint32_t color, int simultaneous, int cycles, int speed) {
  int flashes[simultaneous];

  for (int i = 0; i < cycles; i++) {
    for (int j = 0; j < simultaneous; j++) {
      yield();
      int idx = random(NUM_LEDS);
      flashes[j] = idx;
      matrix.setPixelColor(idx, color ? color : randomColor());
    }
    matrix.show();
    delay(speed);
    for (int s = 0; s < simultaneous; s++) {
      matrix.setPixelColor(flashes[s], 0);
    }
    matrix.show();
    delay(speed);
  }
}

// Speed of background noise
const uint16_t speed = 3;

// Scale determines how far apart the pixels in the noise matrix are
const uint16_t scale = 50;

// Array of computed noise values
uint8_t noise[8][8];

void doPlasma(int loops) {

  // Initialize noise coordinates to random values
  uint16_t x = random16();
  uint16_t y = random16();
  uint16_t z = random16();

  // Do loops cycles of the plasma pattern
  for (int l = 0; l < loops; l++) {

    // Fill the noise array of 8-bit noise values using the inoise8 function.
    for (int i = 0; i < 8; i++) {
      int ioffset = scale * i;
      for (int j = 0; j < 8; j++) {
        int joffset = scale * j;
        noise[i][j] = inoise8(x + ioffset, y + joffset, z);
      }
    }
    z += speed;

    for (int i = 0; i < 8; i++) {
      for (int j = 0; j < 8; j++) {
        // We use the value at the (i,j) coordinate in the noise
        // array for our brightness, and the flipped value from (j,i)
        // for our pixel's hue.
        CRGB _color = CRGB(CHSV(noise[j][i], 255, noise[i][j]));
        uint32_t color = Adafruit_NeoPixel::Color(_color.red, _color.green, _color.blue);

        // Offset from pixel 0 in top left
        int offset = j * 8 + i;

        // Draw on top, left, front, right and back faces
        matrix.setPixelColor(offset      , color);
        matrix.setPixelColor(offset +  64, color);
        matrix.setPixelColor(offset + 128, color);
        matrix.setPixelColor(offset + 192, color);
        matrix.setPixelColor(offset + 256, color);
      }
    }
    matrix.show();
    delay(100);
  }
}

#endif
